#!/usr/bin/env python

from neuron import h, gui
from matplotlib import pyplot
import numpy
import math
import multiprocessing
import itertools as it
import run as rn
import time
import json
import os.path as osp
import glob
import os
import importlib.util as impu


def attach_noise_sin_clamp(cell, delay, dur, offset, amp, freq, dt, tau, sigma, mu, loc, seed):
    """Attach a sinusoidal current Clamp to a cell.

    :param cell: Cell object to attach the current clamp.
    :param delay: Onset of the injected current.
    :param dur: Duration of the stimulus.
    :param offset: Offset of the sine.
    :param amp: The amplitude of the sine.
    :param freq: The frequency of the sine.
    :param sigma: The standard deviation of the normrand.
    :param loc: Location on the dendrite where the stimulus is placed.
    """
    stim = h.IClampNoiseSin(cell.soma[0](loc))
    stim.delay = delay
    stim.dur = dur
    stim.std = sigma
    stim.offset = offset
    stim.amp = amp
    stim.freq = freq
    stim.dt = dt
    stim.tau = tau
    stim.mu = mu
    stim.new_seed = seed
    return stim


def simulate(simdur, dt, init, cel):
    """Start the simulation
    :param simdur: The simulation duration
    :param dt: The time step
    :param init: The initiation voltage
    :param cel: The temperature of the simulationn."""
    h.dt = dt
    h.tstop = simdur
    h.v_init = init
    h.celsius = cel
    h.run()


def set_recording_vectors(stim, cell):
    """Creates the recording vectors in following order: soma, axon, injected current, t
    :param stim: The stimulated input to calculate the injected current
    :param soma: The soma that needs to be recorded
    :param axon: The axon that needs to be recorded"""
    soma_v_vec = h.Vector()
    axon_v_vec = h.Vector()
    i_inj_vec = h.Vector()
    t_vec = h.Vector()
    soma_v_vec.record(cell.soma[0](0.5)._ref_v)
    axon_v_vec.record(cell.axon[0](0.5)._ref_v)
    i_inj_vec.record(stim._ref_i)
    t_vec.record(h._ref_t)
    return soma_v_vec, axon_v_vec, i_inj_vec, t_vec


def writetofile(*arg):
    filename = arg[0]
    temparray = numpy.zeros((len(arg[1]), len(arg)-1))
    maxlength = len(arg)-1
    for i in range(maxlength):
        temp = numpy.array(arg[i + 1])
        temparray[:, i] = temp
    numpy.savetxt(filename, temparray)


def simulatepar(delay, dur, std, offset, amp, freq, dt, tau, mu, loc, simdur, init, cel, seed, loop, dist, distsec, desMfr, synapse):
    """
    :param delay: The delay of the input current
    :param dur: The duration of the current injection
    :param std: The std of the input noise
    :param offset: The DC current to obtain a firing rate of desMfr
    :param amp: The amplitude of the input sine
    :param freq: The frequency of the input sine
    :param dt: The time resolution of the simulation
    :param tau: The time characteristic of the noise
    :param mu: The mean of the noise
    :param loc: The location of the current injection
    :param simdur: The total simulation duration
    :param init: The initial membrane voltage of the neuron
    :param cel: The temperature of the simulation
    :param seed: The seed for the random number generator
    :param loop: The current loop, for internal use
    :param dist: The segment of the AIS
    :param distsec: The section of the AIS
    :param desMfr: The required firing rate of the neuron model
    :param synapse: Boolean to indicate whether or not to include synapses
    """
    h.cvode.cache_efficient(1)

    cell = createModel(synapse, distsec, dist)

    counts = h.Vector()
    apc = h.APCount(cell.soma[0](0.5))
    apc.thresh = 0
    apc.record(counts)

    stim = attach_noise_sin_clamp(cell, delay, dur, offset, amp, freq, dt, tau, std, mu, loc, seed)
    h.dt = dt
    h.tstop = simdur
    h.v_init = init
    h.celsius = cel
    h.run()
    filename = './OUTPUT' + str(distsec) + 'AISSeg' + str(dist) + 'AISLoc/' + str(seed) + 'spiketrain' + str(desMfr) + 'Hzmfr' + str(std)+ 'std' + str(freq) + 'freq' + str(loop) + 'loop.txt'
    writetofile(filename, counts)
    cell = None
    return 1


def simulatepar_one_arg(a):
    return simulatepar(*a)


def realmainOneLoop(numprocs, initval, freqRange, stdVal, distsec, dist, desMfr, Erest, synapse):
    """
    :param numprocs: The number of parallel processes to be used
    :param initval: The DC current to obtain a firing rate of desMfr
    :param freqRange: The frequency range that needs to be investigated
    :param stdVal: The std of the input noise
    :param distsec: The segment of the AIS
    :param dist: The section of the AIS
    :param desMfr: The required firing rate of the neuron model
    :param Erest: The resting potential
    :param synapse: Boolean to indicate whether or not to include synapses
    """
    p = multiprocessing.Pool(processes=numprocs)
    dt = 0.005
    tau = 10.
    mu = 0.
    loc = 0.5
    simdur = 100000.
    cel = 37
    init = Erest
    amp = stdVal * 0.35
    curr = initval
    delay = 100.
    dur = simdur

    for i in range(0, int(numpy.ceil(len(freqRange)/28.)), 1):
        rangeStart = i*28
        rangeEnd = (i+1)*28
        seedrange = range(rangeStart, rangeEnd, 1)
        if rangeEnd > len(freqRange):
            rangeEnd = None
            seedrange = range(rangeStart, len(freqRange), 1)
        frange = freqRange[rangeStart:rangeEnd]

        q = p.map_async(simulatepar_one_arg,
                        zip(it.repeat(delay), it.repeat(dur), it.repeat(stdVal), it.repeat(curr), it.repeat(amp),
                            frange, it.repeat(dt), it.repeat(tau), it.repeat(mu), it.repeat(loc),
                            it.repeat(simdur), it.repeat(init), it.repeat(cel), seedrange, it.repeat(1), it.repeat(dist),
                            it.repeat(distsec), it.repeat(desMfr), it.repeat(synapse)))
        while not q.ready():
            time.sleep(1)


def findStd(cell, desStd):
    """
    :param cell: The neuron model
    :param desStd: The required std value of subthreshold membrane fluctuations
    """
    delay = 0
    dur = 10000
    offset = 0
    amp = 0
    freq = 0
    dt = 0.005
    tau = 10
    sigma = 0
    mu = 0
    loc = 0.5
    seed = 1
    stim = attach_noise_sin_clamp(cell, delay, dur, offset, amp, freq, dt, tau, sigma, mu, loc, seed)

    # Setting up recording vectors
    somav = h.Vector()
    somav.record(cell.soma[0](0.5)._ref_v)
    timevec = h.Vector()
    timevec.record(h._ref_t)
    iinj = h.Vector()
    iinj.record(stim._ref_i)
    # Find input resistance and resting potential
    Erest, Rin = findInputRes(stim, somav, timevec, iinj)

    stim.offset = (-75-Erest)/Rin # To have membrane potential at -75 mV

    # Find right std value of the noise
    h.v_init = Erest
    h.tstop = 2000
    stdVal = 0.001
    errorstd = 1  # To initialize the while loop
    stim.std = stdVal  # To initialize the while loop
    stdlow = 0
    stdhigh = 15/Rin
    somastdlow = 0
    somastdhigh = 30
    h.run()
    somaStd = somav.stdev(0, somav.size() - 1)
    somav.remove(0, somav.size() - 1)
    timevec.remove(0, timevec.size() - 1)
    iinj.remove(0, iinj.size() - 1)
    counter = 0
    while errorstd > 0.01:
        if desStd > somaStd:
            stdhigh = stdVal
            somastdhigh = somaStd
        else:
            stdlow = stdVal
            somastdlow = somaStd

        stdVal = (stdlow * (somastdhigh - desStd) - stdhigh * (somastdlow - desStd)) / (somastdhigh - somastdlow)
        stim.std = stdVal
        h.run()
        somaStd = somav.stdev(0, somav.size() - 1)
        errorstd = numpy.abs(desStd - somaStd)

        somav.remove(0, somav.size() - 1)
        timevec.remove(0, timevec.size() - 1)
        iinj.remove(0, iinj.size() - 1)
        counter += 1
        if counter > 500:
            exit("Couldn't find suitable std")

    strtoprint = "The input resistance is: " + str(Rin) + ", the resting potential is: " + str(Erest) + ", the standard deviation of the noise is set to " + str(stdVal) + "."
    print(strtoprint)
    return stdVal, Erest, Rin


def findInputRes(stim, somav, timevec, iinj):
    """
    :param stim: A NEURON stimulus electrode
    :param somav: A vector to save somatic membrane voltage
    :param timevec: A vector to save the timings of the simulations in
    :param iinj: A vector to save the injected current in
    """
    tstop = 1000
    simulate(tstop, 0.005, -70, 37)
    Erest = somav.x[-2]
    h.tstop = 500
    Rin = numpy.zeros((11, 1))
    count = 0
    for i in numpy.arange(-0.02, -0.2, -0.18/10):
        stim.offset = i
        somav.remove(0, somav.size()-1)
        timevec.remove(0, timevec.size()-1)
        iinj.remove(0, iinj.size()-1)
        h.run()
        Rin[count] = (somav.x[-1]-Erest)/i
        count += 1
    somav.remove(0, somav.size() - 1)
    timevec.remove(0, timevec.size() - 1)
    iinj.remove(0, iinj.size() - 1)
    return Erest, numpy.mean(Rin)


def findMfr(cell, stdVal, Erest, Rin, desMfr):
    """
    :param cell: The neuron model
    :param stdVal: The required std value of subthreshold membrane fluctuations
    :param Erest: The resting potential
    :param Rin: The input resistance
    :param desMfr: The required firing rate of the neuron model
    """
    # Setting up an initial stimulator
    delay = 100
    dur = 1000000
    offset = 0.
    amp = 0
    freq = 0
    dt = 0.005
    tau = 10
    sigma = stdVal
    mu = 0
    loc = 0.5
    seed = 1
    stim = attach_noise_sin_clamp(cell, delay, dur, offset, amp, freq, dt, tau, sigma, mu, loc, seed)

    h.v_init = Erest
    h.tstop = 200000

    # Setting up the AP vector
    counts = h.Vector()
    apc = h.APCount(cell.soma[0](0.5))
    apc.thresh = 0
    apc.record(counts)

    # Loop parameters
    errormfr = 1
    offsetlow = -1
    offsethigh = (-Erest)/Rin
    mfrlow = 0
    mfrhigh = desMfr + 10
    h.run()
    if len(counts) > 2:
        currmfr = 1000 * len(counts) / (counts[-1] - counts[0])
        counts.remove(0, counts.size() - 1)
    else:
        currmfr = 0
    counter = 0
    while errormfr > 0.2:
        if currmfr > desMfr:
            offsethigh = offset
            mfrhigh = currmfr
        else:
            offsetlow = offset
            mfrlow = currmfr

        offset = (offsetlow * (mfrhigh - desMfr) - offsethigh * (mfrlow - desMfr)) / (mfrhigh - mfrlow)
        stim.offset = offset
        h.run()
        if len(counts) > 2:
            currmfr = 1000*len(counts)/(counts[-1]-counts[0])
            counts.remove(0, counts.size() - 1)
        else:
            currmfr = 0

        errormfr = abs(desMfr - currmfr)
        counter += 1
        if counter > 50:
            exit("Couldn't find a suitable offset for the desired mfr.")

    strtoprint = "The mean of the input is set to: " + str(offset) + "."
    print(strtoprint)
    return offset


def jsondump(Erest, Rin, stdVal, somastd, offset, currmfr, aisSeg, aisLoc, filenam):
    """
    :param Erest: The resting potential
    :param Rin: Input resistance
    :param stdVal: The std value of the input noise
    :param somaStd: The required std value of subthreshold membrane fluctuations
    :param offset: The DC current required to obtain currmfr as a firing rate
    :param currmfr: The required firing rate of the neuron
    :param aisSeg: Segment of the location of the AIS
    :param aisLoc: Section of the location of the AIS
    :param filenam: The filename of the file to save all the data
    """
    parameters = {"Erest": Erest,
                  "Rin": Rin,
                  "stdVal": stdVal,
                  "somaStd": somastd,
                  "offset": offset,
                  "mfr": currmfr,
                  "aisSeg": aisSeg,
                  "aisloc": aisLoc,
                  }
    fid = open(filenam, "w")
    json.dump(parameters, fid)
    fid.close()


def runStdMfrProtocol(cell, aisSeg, aisLoc, desStd, desMfr):
    """
    :param synapse: Boolean to indicate if synapses are included
    :param aisSeg: Segment of the location of the AIS
    :param aisLoc: Section of the location of the AIS
    :param desStd: The required std value of subthreshold membrane fluctuations
    :param desMfr: The required firing rate of the neuron
    """
    filenam = "./Parameters/" + str(aisSeg) + "aisSeg" + str(aisLoc) + "aisloc" + str(desStd) + "SomaStd" + str(desMfr) + "Hz.json"

    if osp.isfile(filenam):
        fid = open(filenam, "r")
        tempJsonVar = json.load(fid)
        fid.close()

        Erest = tempJsonVar["Erest"]
        Rin = tempJsonVar["Rin"]
        stdVal = tempJsonVar["stdVal"]
        somaStd = tempJsonVar["somaStd"]
        offset = tempJsonVar["offset"]
        mfr = tempJsonVar["mfr"]
        aisLoc = tempJsonVar["aisloc"]
        print("Found existing JSON file for the requested paramters. Opening existing JSON file and using these input parameters: \n - Erest: " + str(Erest) + "\n - Rin:" + str(Rin) + "\n - stdVal: " + str(stdVal) + "\n - somaStd: " + str(somaStd) + "\n - offset: " + str(offset) + "\n - mfr: " + str(mfr) + "\n - aisloc: " + str(aisLoc))
    else:
        print("No JSON file detected for required parameters, running the following protocol: \n - Mcm.findInputRes \n - Mcm.findStd \n - Mcm.findMfr \n - Mcm.jsondump")

        # Finding the right std for the noise
        stdVal, Erest, Rin = findStd(cell, desStd)

        # Finding the correct offset for the desired mfr
        offset = findMfr(cell, stdVal, Erest, Rin, desMfr)

        # Saving the found parameters in a json file
        jsondump(Erest, Rin, stdVal, desStd, offset, desMfr, aisSeg, aisLoc, filenam)

    return Erest, Rin, stdVal, offset


def createModel(synapse, aisSeg, aisLoc):
    """
    :param synapse: Boolean to indicate if synapses are included
    :param aisSeg: Segment of the location of the AIS
    :param aisLoc: Section of the location of the AIS
    """
    rn.init_simulation()
    cell = rn.create_cell(synapse)

    ### Delete this part once changing AIS is no longer needed
    orisodium_t = cell.axon[1].gNaTa_tbar_NaTa_t
    orisodium_p = cell.axon[1].gNap_Et2bar_Nap_Et2
    oripot_t = cell.axon[1].gK_Tstbar_K_Tst
    oripot_p = cell.axon[1].gK_Pstbar_K_Pst

    cell.axon[0].nseg = 21
    cell.axon[1].nseg = 21
    cell.axon[0].gNaTa_tbar_NaTa_t = orisodium_t * 1
    cell.axon[1].gNaTa_tbar_NaTa_t = orisodium_t * 1
    cell.axon[0].gK_Tstbar_K_Tst = oripot_t * 1
    cell.axon[1].gK_Tstbar_K_Tst = oripot_t * 1
    cell.axon[aisSeg](aisLoc).gNaTa_tbar_NaTa_t = orisodium_t * 18.
    cell.axon[aisSeg](aisLoc).gK_Tstbar_K_Tst = oripot_t * 18.

    cell.axon[0].gNap_Et2bar_Nap_Et2 = orisodium_p * 1
    cell.axon[1].gNap_Et2bar_Nap_Et2 = orisodium_p * 1
    cell.axon[0].gK_Pstbar_K_Pst = oripot_p * 1
    cell.axon[1].gK_Pstbar_K_Pst = oripot_p * 1
    cell.axon[aisSeg](aisLoc).gNap_Et2bar_Nap_Et2 = orisodium_p * 18
    cell.axon[aisSeg](aisLoc).gK_Pstbar_K_Pst = oripot_p * 18

    return cell
