#!/usr/bin/env python3
#
# figures.py
#

import matplotlib.pyplot as plt
import nest
import numpy as np
from os.path import join
import scipy.stats

from utils.misc import setup_numpy_and_matplotlib
from utils.plot import grouped_barplot, barplot_annotate_brackets, inset_text


def plot_comparison_seeds(training_data_, labels_, label, title=None, show_boxplot=False, outdir=None):
    setup_numpy_and_matplotlib(plt_size=11)

    training_data_m_ = []

    for k, td in enumerate(training_data_):
        td = [np.mean(x) for x in training_data_[k]]
        training_data_m_ += [td]

    # pool data

    for k in range(len(training_data_)):
        training_data_[k] = [y for x in training_data_[k] for y in x]

    # process

    def process(training_data_, labels_, label, title, suffix, outdir):

        x_ = np.arange(len(training_data_))

        means = [np.mean(td) for td in training_data_]
        std = [np.std(td) for td in training_data_]
        sem = [np.std(td)/np.sqrt(len(td)) for td in training_data_]
        dy = max(means) - min(means)
        y0 = min(means) - .4*dy
        y1 = max(means) + dy

        for l, td, m, s in zip(x_, training_data_, means, std):
            print('#{0:s}: N = {1:d}, mean = {2:.3f}, std = {3:.3f}'.format(str(l), len(td), m, s))

        pval_ = []
        for j, k in zip(range(0, len(training_data_)-1), range(1, len(training_data_))):
            pval = scipy.stats.ranksums(training_data_[j], training_data_[k])[1]

            print('ranksum td[{0:d}] and td[{1:d}]: {2:.5f}'.format(j, k, pval))

            pval_ += [pval]

        pval_all = scipy.stats.ranksums(training_data_[0], training_data_[-1])[1]
        print('ranksum td[0] and td[-1]: {0:.5f}'.format(pval_all))

        # plot all pvals
        plot_pvals = [(k, k+1, pval) for k, pval in enumerate(pval_)]

        plt.figure(figsize=(12 if show_boxplot else 6, 4))

        # boxplot

        if show_boxplot:
            plt.subplot(1, 2, 1)
            plt.boxplot(training_data_)
            plt.gca().set_xticks(labels_[0]+1)
            plt.gca().set_xticklabels(labels_[1])
            plt.xlabel(label)
            plt.ylabel('performance')
            plt.locator_params('y', nbins=5)
            if title:
                plt.title(title)

        # bars with p-values

        tops = [m+s for m, s in zip(means, std)]
        ymin = min([m-s for m, s in zip(means, std)])
        ymax= max([m+s for m, s in zip(means, std)])
        y0 = ymin - (ymax - ymin) * .1
        y1 = ymax + (ymax - ymin) * .5

        if show_boxplot:
            plt.subplot(1, 2, 2)

        plt.bar(x_, means, yerr=std, align='center', ecolor='black', capsize=10., color='C2')
        plt.gca().set_xticks(labels_[0])
        plt.gca().set_xticklabels(labels_[1])
        plt.xlabel(label)
        plt.ylim(y0, y1)
        plt.ylabel('performance')
        plt.locator_params('y', nbins=5)
        yticks = plt.gca().get_yticks()
        plt.gca().set_yticklabels(['{0:.1f}%'.format(yt*100) for yt in yticks])

        bar0, bar1, bar2 = 0, round(len(training_data_)/2)-1, len(training_data_)-1
        for bar1, bar2, pval in plot_pvals:
            barplot_annotate_brackets(bar1, bar2, pval, x_, tops, fs=12, maxasterix=5)
        barplot_annotate_brackets(bar0, bar2, pval_all, x_, tops, dh=.15, fs=12, maxasterix=5)
        inset_text('(N = {0:d} each)'.format(len(training_data_[0])), 'top', 'right', fontsize=10)
        if title:
            plt.title(title)

        plt.tight_layout()

        if outdir:
            plt.savefig(join(outdir, 'plot'+suffix+'_all.pdf'))

    process(training_data_m_, labels_, label, title+'', '_means', outdir=outdir)
