#!/usr/bin/env python
#
# main_lsm_multirun.py
#

import argparse
import matplotlib as mpl
mpl.use('Agg')
import matplotlib.pyplot as plt
from os.path import join
import numpy as np
from timeit import default_timer as timer

from utils.misc import projected_time_left
import utils.nest  # (sort of) silent nest import
from utils.simulation import Simulation

from config import read_config
from model import lsm_experiment_run, lsm_experiment_train


if __name__ == '__main__':
    parser = argparse.ArgumentParser(description='Generate LSM experiment data.')
    parser.add_argument('-s', '--scenario', dest='scenario', choices=['opt_for_25um', 'opt_for_DeltaT2.5', 'opt_for_equal_delays', 'opt_for_0um'], default='opt_for_25_um', help='type of data to generate', required=True)
    parser.add_argument('-n', '--num-seeds', dest='num_seeds', type=int, default=50, help='number of random seeds, default: 50')

    args = parser.parse_args()

    task = 'multi'
    simtime = 500e3
    scenario = args.scenario

    seed_ = list(range(1, args.num_seeds+1))

    if scenario == 'opt_for_25um':
        network_cfg = 'network_opt_for_25um'

        # eif models fit to bas position
        neuron_cfg_ = ['neuron_bas_pos{0:02d}'.format(n) for n in range(11)]

    elif scenario == 'opt_for_DeltaT2.5':
        network_cfg = 'network_opt_for_DeltaT2.5'

        # data from eif model fit to 25 um, just varying Delta T
        neuron_cfg_ = ['neuron_basm_DeltaT{0:.1f}'.format(Delta_T) for Delta_T in [0.5, 2.5, 4.5]]

    elif scenario == 'opt_for_equal_delays':
        network_cfg = 'network_opt_for_equal_delays'

        # eif models fit to bas position
        neuron_cfg_ = ['neuron_bas_pos{0:02d}'.format(n) for n in range(11)]

    elif scenario == 'opt_for_0um':
        network_cfg = 'network_opt_for_0um'

        # eif models fit to bas position
        neuron_cfg_ = ['neuron_bas_pos{0:02d}'.format(n) for n in range(11)]

    else:
        raise ValueError()


    # read config files
    network_cfg = read_config(network_cfg)

    sim_name = join('_'.join([task, network_cfg.name, *neuron_cfg_[0].split('_')[:2]]))

    # run

    steps_total = len(seed_) * len(neuron_cfg_)
    step = 0
    step_durations = []

    for seed in seed_:
        for neuron_cfg in neuron_cfg_:
            neuron_cfg = read_config(neuron_cfg)
            suffix = '_'.join([task, neuron_cfg.name, network_cfg.name, 'seed'+str(seed)])

            t0 = timer()

            with Simulation(sim_name=sim_name, suffix=suffix, numbered=False) as sim:
                data = lsm_experiment_run(
                        task,
                        neuron_cfg=neuron_cfg,
                        network_cfg=network_cfg,
                        sim_time=simtime,
                        seed=seed,
                        sim=sim,
                        plot=True)

                lsm_experiment_train(data, sim=sim, plot=True)

            plt.close('all')

            t1 = timer()

            step += 1
            step_durations += [t1-t0]

            print('runs done: {0:d}/{1:d}  projected time left: {2:s}'.format(
                step, steps_total, projected_time_left(steps_total, step, step_durations)))
