#!/usr/bin/env python
#
# main_plot_results.py
#

import argparse
import glob
import matplotlib as mpl
#mpl.use('Agg')
import matplotlib.pyplot as plt
import os.path
from os.path import join
import numpy as np

from utils.simulation import Simulation
import utils.nest  # (sort of) silent nest import

from config import read_config
from figures import plot_comparison_seeds
from model import lsm_experiment_run, lsm_experiment_train


if __name__ == '__main__':
    parser = argparse.ArgumentParser(description='Generate LSM experiment data.')
    parser.add_argument('-s', '--scenario', dest='scenario', choices=['opt_for_25um', 'opt_for_DeltaT2.5', 'opt_for_equal_delays', 'opt_for_0um'], default='opt_for_25_um', help='type of data to generate', required=True)

    args = parser.parse_args()

    task = 'multi'
    scenario = args.scenario
    train = False

    if scenario == 'opt_for_25um':
        network_cfg = 'network_opt_for_25um'

        # eif models fit to bas position
        neuron_cfg_ = ['neuron_bas_pos{0:02d}'.format(n) for n in range(11)]

        title = r'BAS opt. for $25\,\mathrm{\mu m}$'

    elif scenario == 'opt_for_DeltaT2.5':
        network_cfg = 'network_opt_for_DeltaT2.5'

        # data from eif model fit to 25 um, just varying Delta T
        neuron_cfg_ = ['neuron_basm_DeltaT{0:.1f}'.format(Delta_T) for Delta_T in [0.5, 2.5, 4.5]]

        title = r'BAS middle pos., opt. for $\Delta t = 2.5\,\mathrm{mV}$'

    elif scenario == 'opt_for_equal_delays':
        network_cfg = 'network_opt_for_equal_delays'

        # eif models fit to bas position
        neuron_cfg_ = ['neuron_bas_pos{0:02d}'.format(n) for n in range(11)]

        title = r'BAS (equal delays) opt. for $25\,\mathrm{\mu m}$'

    elif scenario == 'opt_for_0um':
        network_cfg = 'network_opt_for_0um'

        # eif models fit to bas position
        neuron_cfg_ = ['neuron_bas_pos{0:02d}'.format(n) for n in range(11)]

        title = r'BAS opt. for $0\,\mathrm{\mu m}$'

    else:
        raise ValueError()

    # ----------------------------------------------------------------------
    # setup

    network_cfg = read_config(network_cfg)
    sim_name = join('_'.join([task, network_cfg.name, *neuron_cfg_[0].split('_')[:2]]))

    training_data_ = []

    for neuron_cfg in neuron_cfg_:
        neuron_cfg = read_config(neuron_cfg)

        base = '_'.join([task, neuron_cfg.name, network_cfg.name, ''])
        suffix_ = glob.glob(join('out', sim_name, base+'*'))
        suffix_ = [os.path.basename(suffix) for suffix in suffix_]

        training_data_ += [[]]

        for k, suffix in enumerate(suffix_):

            if train:
                with Simulation(sim_name=sim_name, suffix=suffix, numbered=False, suspend=True) as sim:

                    data = sim.restore_data()

                    cost, training_data = lsm_experiment_train(data, sim=sim, plot=True)
            else:
                try:
                    with Simulation(sim_name=sim_name, suffix=suffix, numbered=False, suspend=True) as sim:
                        training_data = sim.restore_data(title='training')
                except IOError as err:
                    print('cannot use {0:s} ({1:s}), skipping'.format(suffix, str(err)))
                    continue

            plt.close('all')

            if 'test_error_all' not in training_data[sorted(training_data.keys())[0]]:
                print('cannot use '+ suffix + ', skipping')
                continue

            # ignoring all but xor for now

            assert 'xor' in training_data.keys()

            key = 'xor'
            reg_, reg = training_data[key]['reg_'], training_data[key]['reg']
            ind = np.where(reg_ == reg)[0][0]

            training_data_[-1] += [(1-training_data['xor']['test_error_all'][ind]).tolist()]

    if len(np.unique([len(td) for td in training_data_])) > 1:
        print('incosistent lengths, cropping')
        L = np.unique([len(td) for td in training_data_]).min()
        training_data_ = [td[:L] for td in training_data_]

    label_ = [nc.split('_')[2] for nc in neuron_cfg_]
    if 'DeltaT' in label_[0]:
        label_ = [float(l[6:]) for l in label_]
        label = r'$\Delta_\mathrm{T}$ [mV]'
        labels = (np.arange(len(label_)), label_)
    else:
        label = r'soma-AIS distance [$\mu$m]'
        labels = (np.asarray([0, 2, 4, 6, 8, 10]), [0, 10, 20, 30, 40, 50])

    plot_comparison_seeds(training_data_, labels, label, title, outdir=join('out', sim_name))

    plt.show()
