#!/usr/bin/env python3
#
# model.py
#

import matplotlib.pyplot as plt
import multiprocessing as mp
import nest
import numpy as np
from os.path import join
import pickle as pkl
from sklearn.linear_model import LogisticRegression
from sklearn.preprocessing import StandardScaler
from sklearn.model_selection import train_test_split

from utils.misc import *
from utils.nest import *
from utils.spikes import get_liquid_states

from stimulus import *


def linreg_readout(X, y, Xt, yt, reg_fact=0):
    """
    Train and test a linear regression readout.

    inputs:
        X               training features
        y               training targets
        Xt              test features
        yt              test targets
        reg_fact        regularization factor

    returns:
        train_err       error on training data
        test_err        error on test data
    """
    # add constant component to states for bias
    X = np.hstack((np.ones((len(X), 1)), X))
    Xt = np.hstack((np.ones((len(Xt), 1)), Xt))

    # train
    if reg_fact == 0:
        w = np.linalg.lstsq(X, y)[0]
    else:
        I = np.eye(X.shape[1])
        w = np.linalg.inv(np.dot(X.T, X) + reg_fact*I).dot(X.T).dot(y)

    # compute predictions
    y_train = (np.dot(X, w) >= .5).astype(int)
    y_test = (np.dot(Xt, w) >= .5).astype(int)

    # compute errors
    train_err = (y_train != y).sum() / len(y)
    test_err = (y_test != yt).sum() / len(yt)

    return train_err, test_err


def logreg_readout(X, y, Xt, yt, reg_fact=0.):
    """
    Train and test a logistic regression readout. If y and yt do not
    have dtype bool, > 0. will be applied.

    inputs:
        X               training features
        y               training targets
        Xt              test features
        yt              test targets
        reg_fact        regularization factor

    returns:
        train_err       error on training data
        test_err        error on test data
    """
    scaler = StandardScaler()
    x_scaled = scaler.fit_transform(X)
    xt_scaled = scaler.transform(Xt)

    C = 1. / (reg_fact + 1.)

    net = LogisticRegression(
            solver='sag',
            max_iter=2000,
            C=C,
            n_jobs=mp.cpu_count())

    if not y.dtype is np.dtype(bool):
        y = y > 0.

    if not yt.dtype is np.dtype(bool):
        yt = yt > 0.

    net.fit(X, y)

    # training error
    y_pred = net.predict(X)
    mcr_train = sum(y_pred != y) / len(y)

    # test error
    yt_pred = net.predict(Xt)
    mcr_test = sum(yt_pred != yt) / len(yt)

    return mcr_train, mcr_test


def train_readouts(states, targets, reg_, num_trained_classifiers=100, logreg=False, verbose=False):
    """
    Train and test readouts for many regularization values and report
    the best error.

    inputs:
        states          liquid states, will be split 80/20 for
                        training/test
        targets         targets
        reg_            list of regularization factors to try
        num_trained_classifiers
                        number of classifiers to train with random
                        training/test splits
        logreg          train a logistic regression readout, if
                        False, train a linear regression readout
        verbose         be verbose

    returns:
        results         dict containing results
    """
    train_mean_, train_std_ = [], []
    test_mean_, test_std_ = [], []
    test_mean_all_ = []

    for reg in reg_:
        if verbose:
            print('  using regularization = {0:e}'.format(reg))

        train_mean_cur_ = []
        test_mean_cur_ = []

        for i in range(num_trained_classifiers):
            states_train, states_test, targets_train, targets_test = train_test_split(
                    states, targets, train_size=.8, test_size=.2)

            if not logreg:
                train_mean, test_mean = linreg_readout(
                        states_train, targets_train, states_test, targets_test, reg_fact=reg)
            else:
                train_mean, test_mean = logreg_readout(
                        states_train, targets_train, states_test, targets_test, reg_fact=reg)

            train_mean_cur_ += [train_mean]
            test_mean_cur_ += [test_mean]

        train_mean, train_std = np.mean(train_mean_cur_), np.std(train_mean_cur_)
        test_mean, test_std = np.mean(test_mean_cur_), np.std(test_mean_cur_)

        train_mean_ += [train_mean]
        train_std_ += [train_std]
        test_mean_ += [test_mean]
        test_std_ += [test_std]
        test_mean_all_ += [np.asarray(test_mean_cur_)]

        if verbose:
            print('    train error mean: {0:.3f}, std: {1:.3f}'.format(train_mean, train_std))
            print('    test error mean: {0:.3f}, std: {1:.3f}'.format(test_mean, test_std))

    train_mean_, train_std_ = np.asarray(train_mean_), np.asarray(train_std_)
    test_mean_, test_std_ = np.asarray(test_mean_), np.asarray(test_std_)

    ind_min = test_mean_.argmin()

    results = {
            'num_trained_classifiers': num_trained_classifiers,
            'logreg': logreg,
            'reg_': reg_,
            'train_errors': [train_mean_, train_std_],
            'test_errors': [test_mean_, test_std_],
            'reg': reg_[ind_min],
            'test_error_mean': test_mean_[ind_min],
            'test_error_std': test_std_[ind_min],
            'test_error_all': test_mean_all_}

    return results


def plot_single_delay(training_data):
    """
    Plot results for classification task.

    inputs:
        training_data
    """
    train_mean_, train_std_ = training_data['train_errors']
    test_mean_, test_std_ = training_data['test_errors']
    train_at_min = train_mean_[test_mean_.argmin()]
    reg = training_data['reg']
    reg_ = training_data['reg_']

    c = 'C0'

    plt.figure(figsize=(12., 4.))

    plt.subplot(1, 2, 1)
    plt.plot(np.log(reg_), train_mean_, c=c, lw=2)
    plt.fill_between(np.log(reg_), train_mean_+train_std_, train_mean_-train_std_, color=c, alpha=.5)
    plt.xlabel(r'$\ln \lambda$')
    plt.ylabel(r'training error')
    plt.title(r'{0:.2f} with $\lambda={1:.1g}$'.format(train_at_min, reg))

    plt.subplot(1, 2, 2)
    plt.plot(np.log(reg_), test_mean_, c=c, lw=2)
    plt.fill_between(np.log(reg_), test_mean_+test_std_, test_mean_-test_std_, color=c, alpha=.5)
    plt.xlabel(r'$\ln \lambda$')
    plt.ylabel(r'test error')
    plt.title(r'min: {0:.2f} with $\lambda={1:.1g}$'.format(test_mean_.min(), reg))

    plt.tight_layout()


def plot_multi_delay(training_data):
    """
    Plot results for pattern memory task.

    inputs:
        training_data
    """
    delay_, test_mean_ = [], []

    for delay, r in sorted(training_data.items()):
        test_mean_ += [r['test_errors'][0].min()]
        delay_ += [delay]

    plt.figure(figsize=(6., 4.))

    plt.bar(delay_[::-1], 1.-np.asarray(test_mean_))
    plt.xlabel('delay')
    plt.ylabel('accuracy')
    plt.xticks(delay_[::-1], delay_)

    plt.tight_layout()


# ----------------------------------------------------------------------
# main

def lsm_experiment_run(task, neuron_cfg, network_cfg, sim_time=200e3, seed=None, sim=None, plot=False):
    """
    Run network for LSM experiment.

    inputs:
        task            name of task
        neuron_cfg      neuron parameters
        network_cfg     network parameters
        sim_time        simulation time
        seed            seed for randomness
        sim             simulation instance
        plot            whether to generate plots

    returns:
        results         dict containing results
    """
    dump_dict(neuron_cfg, key='neuron_cfg')
    dump_dict(network_cfg, key='network_cfg')

    # setup files and output directory

    if sim:
        pklfile_sim = join(sim.outdir, 'data_sim_{0:s}.pkl'.format(task))
        pklfile_train = join(sim.outdir, 'data_train_{0:s}.pkl'.format(task))

    # setup libs

    setup_numpy_and_matplotlib(seed=seed)
    setup_nest(modname='unsafemodule', thread_offs=0, thread_div=1., print_time=False, seeds=seed)

    dt = nest.GetKernelStatus('resolution')

    # ----------------------------------------------------------------------
    # create nodes

    neuron_params = {
            'C_m': neuron_cfg.C_m,
            'V_reset': neuron_cfg.V_reset,
            'V_th': neuron_cfg.V_T,
            'E_L': neuron_cfg.E_L,
            'V_m': neuron_cfg.E_L,
            'g_L': neuron_cfg.C_m / neuron_cfg.tau_m,
            'I_e': 0.,
            'Delta_T': neuron_cfg.Delta_T,
            'tau_syn_ex': network_cfg.tau_syn_E,
            'tau_syn_in': network_cfg.tau_syn_I,
            'tau_syn_in': network_cfg.tau_syn_I,
            't_ref': neuron_cfg.t_ref,
            'a': 0.,
            'b': 0.}

    E_population = nest.Create('eif_psc_exp', network_cfg.N_E, neuron_params)
    I_population = nest.Create('eif_psc_exp', network_cfg.N_I, neuron_params)

    population = E_population + I_population

    noise_generator = nest.Create('poisson_generator', params={'rate': network_cfg.rho_noise})

    # create spike detectors from excitatory and inhibitory populations
    in_spike_detector = nest.Create('spike_detector')
    E_spike_detector = nest.Create('spike_detector')
    I_spike_detector = nest.Create('spike_detector')

    # generate input spikes

    if task == 'none':
        input_spikes, targets = [[]], []
        input_params = {}

    elif task == 'multi':
        num_pattern = 2
        num_inputs = 10
        stim_len = 50.  # ms
        rate = 20.  # Hz

        targets, templates, input_spikes = generate_stimulus_spatiotemp_multitarget(
                num_pattern=num_pattern,
                num_inputs=num_inputs,
                stim_len=stim_len,
                t_sim=sim_time,
                rate=rate)

        input_params = {
                'num_pattern': num_pattern,
                'num_inputs': num_inputs,
                'stim_len': stim_len,
                'rate': rate,
                'templates': templates}

    else:
        raise ValueError()

    # create spike generators
    params = [{'spike_times': sp} for sp in input_spikes]
    spike_generators = nest.Create('spike_generator', len(params), params=params)

    # ----------------------------------------------------------------------
    # connect nodes

    delay_dict_E = dict(
            distribution='uniform',
            low=network_cfg.synd_E_low,
            high=network_cfg.synd_E_high)
    delay_dict_I = dict(
            distribution='uniform',
            low=network_cfg.synd_I_low,
            high=network_cfg.synd_I_high)

    syn_spec_EE = {
            'model': 'tsodyks_synapse',
            'weight': {
                'distribution': 'normal',
                'mu': network_cfg.w0*network_cfg.J_EE,
                'sigma': .7*network_cfg.w0*network_cfg.J_EE},
            'delay': delay_dict_E,
            'tau_psc': 2.,
            'tau_fac': 1.,
            'tau_rec': 813.,
            'U': .59,
            'u': .59,
            'x': 0.}
    conn_spec_EE = {'rule': 'fixed_indegree', 'indegree': network_cfg.C_E}
    nest.Connect(E_population, E_population, conn_spec_EE, syn_spec_EE)

    syn_spec_EI = {
            'model': 'tsodyks_synapse',
            'weight': {
                'distribution': 'normal',
                'mu': network_cfg.w0*network_cfg.J_EI,
                'sigma': .7*network_cfg.w0*network_cfg.J_EI},
            'delay': delay_dict_E,
            'tau_psc': 2.,
            'tau_fac': 1790.,
            'tau_rec': 399.,
            'U': .049,
            'u': .049,
            'x': 0.}
    conn_spec_EI = {'rule': 'fixed_indegree', 'indegree': network_cfg.C_E}
    nest.Connect(E_population, I_population, conn_spec_EI, syn_spec_EI)

    syn_spec_IE = {
            'model': 'tsodyks_synapse',
            'weight': {
                'distribution': 'normal',
                'mu': -abs(network_cfg.w0*network_cfg.J_IE),
                'sigma': .7*abs(network_cfg.w0*network_cfg.J_IE)},
            'delay': delay_dict_I,
            'tau_psc': 2.,
            'tau_fac': 376.,
            'tau_rec': 45.,
            'U': .016,
            'u': .016,
            'x': 0.}
    conn_spec_IE = {'rule': 'fixed_indegree', 'indegree': network_cfg.C_I}
    nest.Connect(I_population, E_population, conn_spec_IE, syn_spec_IE)

    syn_spec_II = {
            'model': 'tsodyks_synapse',
            'weight': {
                'distribution': 'normal',
                'mu': -abs(network_cfg.w0*network_cfg.J_II),
                'sigma': .7*abs(network_cfg.w0*network_cfg.J_II)},
            'delay': delay_dict_I,
            'tau_psc': 2.,
            'tau_fac': 21.,
            'tau_rec': 706.,
            'U': .25,
            'u': .25,
            'x': 0.}
    conn_spec_II = {'rule': 'fixed_indegree', 'indegree': network_cfg.C_I}
    nest.Connect(I_population, I_population, conn_spec_II, syn_spec_II)

    # inputs

    syn_spec_inp = {
            'model': 'static_synapse',
            'weight': {
                'distribution': 'uniform',
                'low': .5*network_cfg.w0*network_cfg.J_XE,
                'high': 1.5*network_cfg.w0*network_cfg.J_XE},
            'delay': delay_dict_E}
    conn_spec_inp = {
            'rule': 'fixed_outdegree',
            'outdegree': network_cfg.C_inp}
    nest.Connect(spike_generators, E_population, conn_spec_inp, syn_spec_inp)

    # noise

    syn_spec_noise = {'weight': network_cfg.w0*network_cfg.J_noise}
    nest.Connect(noise_generator, population, syn_spec=syn_spec_noise)

    # spike detectors

    nest.Connect(spike_generators, in_spike_detector)
    nest.Connect(E_population, E_spike_detector)
    nest.Connect(I_population, I_spike_detector)

    # voltmeters

    N_vm = 3  # number of neurons to record from
    vm_ = [nest.Create('voltmeter') for k in range(N_vm)]
    [nest.Connect(vm, [E_population[k]]) for k, vm in enumerate(vm_)]

    # ----------------------------------------------------------------------
    # simulate

    nest.Simulate(sim_time)

    # ----------------------------------------------------------------------
    # analysis

    spikes_X = get_spikes_from_spike_recorder(in_spike_detector)
    spikes_E = get_spikes_from_spike_recorder(E_spike_detector)
    spikes_I = get_spikes_from_spike_recorder(I_spike_detector)

    # compute spike rates

    rates_ex = neuron_spike_rates(spikes_E, t_lim=(0., sim_time), pop=E_population)[1]
    rates_in = neuron_spike_rates(spikes_I, t_lim=(0., sim_time), pop=I_population)[1]

    print(('excitatory rate: {0:.2f} +/- {1:.2f} ({2:.2f} -- {3:.2f}) Hz'.format(
        rates_ex.mean(), rates_ex.std(), rates_ex.min(), rates_ex.max())))
    print(('inhibitory rate: {0:.2f} +/- {1:.2f} ({2:.2f} -- {3:.2f}) Hz'.format(
        rates_in.mean(), rates_in.std(), rates_in.min(), rates_in.max())))

    if plot:
        xlim = [0., min(5000., sim_time)]

        plt.figure(figsize=(6, 9))

        plot_spikes(spikes_X, ylabel='X', xlim=xlim, xlabel='', subplot=(4, 1, 1))
        plot_spikes(spikes_E, ylabel='E', xlim=xlim, xlabel='', subplot=(4, 1, 2), targets=E_population[:network_cfg.N_rec], sharex=plt.gca())
        plot_spike_hist(spikes_E, network_cfg.N_E, ylabel='rate in E', xlim=xlim, xlabel='', subplot=(4, 1, 3), sharex=plt.gca())
        plot_spikes(spikes_I, ylabel='I', xlim=xlim, subplot=(4, 1, 4), targets=I_population[:network_cfg.N_rec], sharex=plt.gca())

        plt.tight_layout()

        if sim:
            plt.savefig(join(sim.outdir, 'dynamics_start.pdf'))

        if sim_time > 5000.:
            xlim = [sim_time-5000., sim_time]

            plt.figure(figsize=(6, 9))

            plot_spikes(spikes_X, ylabel='X', xlim=xlim, xlabel='', subplot=(4, 1, 1))
            plot_spikes(spikes_E, ylabel='E', xlim=xlim, xlabel='', subplot=(4, 1, 2), targets=E_population[:network_cfg.N_rec], sharex=plt.gca())
            plot_spike_hist(spikes_E, network_cfg.N_E, ylabel='rate in E', xlim=xlim, xlabel='', subplot=(4, 1, 3), sharex=plt.gca())
            plot_spikes(spikes_I, ylabel='I', xlim=xlim, subplot=(4, 1, 4), targets=I_population[:network_cfg.N_rec], sharex=plt.gca())

            plt.tight_layout()

            if sim:
                plt.savefig(join(sim.outdir, 'dynamics_end.pdf'))

    data = {
            'neuron_cfg': neuron_cfg,
            'network_cfg': network_cfg,
            'task': task,
            'input_params': input_params,
            'sim_params': {'sim_time': sim_time},
            'spikes': [spikes_E, spikes_I, spikes_X],
            'targets': targets,
            'rates': [rates_ex, rates_in]}

    if sim:
        sim.save_data(data)

    return data


def lsm_experiment_train(data, sim=None, plot=False):
    """
    Train readouts for LSM experiment.

    inputs:
        data            data from lsm_experiment_run()
        sim             simulation instance
        plot            whether to generate plots

    returns:
        cost            cost on task
        data            dict containing results
    """

    neuron_cfg = data['neuron_cfg']
    network_cfg = data['network_cfg']
    task = data['task']
    input_params = data['input_params']
    sim_time = data['sim_params']['sim_time']
    spikes_E, spikes_I, spikes_X = data['spikes']
    targets = data['targets']
    rates_ex, rates_in = data['rates']

    # train readouts

    logreg = False

    # training settings
    tau_lsm = 20.  # ms
    reg_ = np.logspace(.01, 15, 50, base=np.e)

    # drop initial responses to allow dynamics to settle
    discard = 5

    print('training...')

    if task == 'multi':
        stim_len = input_params['stim_len']

        training_data = {}

        spike_times = get_spike_times_by_id(spikes_E)[1]

        if len(spike_times) == 0:
            print('no spikes, cannot extract liquid states')
            return np.inf, None

        delay = 0.  # start readout at end of each stimulus
        extract_times = (1+np.arange(0, int(sim_time / stim_len))) * stim_len
        extract_times = np.clip(extract_times+delay, None, sim_time)

        states = get_liquid_states(spike_times[:network_cfg.N_rec], extract_times, tau_lsm)

        # discard first few states
        states = states[discard:,:]

        # classification task
        print('training classification...')

        r = train_readouts(states, targets['classification'][discard:], reg_)

        training_data['classification'] = r
        print('  best: {0:g} (using reg {1:g})'.format(r['test_error_mean'], r['reg']))

        # memory tasks
        training_data['memory'] = {0: r}

        for delay in targets['memory'].keys():
            print('training memory {0:d}...'.format(delay))

            r = train_readouts(states, targets['memory'][delay][discard:], reg_)

            training_data['memory'][delay] = r
            print('  best: {0:g} (using reg {1:g})'.format(r['test_error_mean'], r['reg']))

        # xor task
        print('training xor...')

        r = train_readouts(states, targets['xor'][discard:], reg_)

        training_data['xor'] = r
        print('  best: {0:g} (using reg {1:g})'.format(r['test_error_mean'], r['reg']))

        if plot:
            plot_single_delay(training_data['classification'])

            if sim:
                plt.savefig(join(sim.outdir, 'plot_xor_regs.pdf'))

            plot_multi_delay(training_data['memory'])

            if sim:
                plt.savefig(join(sim.outdir, 'plot_mem_delays.pdf'))

        cost = training_data['xor']['test_error_mean']

    else:
        raise ValueError()

    if sim:
        sim.save_data(training_data, title='training', overwrite=True)

    return cost, training_data
