#!/usr/bin/env python
#
# stimulus.py
#

import numpy as np


def poisson_generator(rate, t_lim, unit_ms=False, round_to=None, seed=None):
    """
    Draw events from a Poisson point process.

    Note: the implementation assumes a spike at t=t_lim[0], although this spike
    is not included in the spike list.

    inputs:
        rate                the rate of the discharge in Hz
        t_lim               tuple containing start and end time
        unit_ms             use ms as unit for times in t_lim and resulting
                            events
        rount_to            if passed, times are rounded to multiples of this
                            number
        seed                seed for random number generation

    returns:
        events              numpy array containing spike times in s (or ms, if
                            unit_ms is set)

    """

    assert len(t_lim) == 2

    if seed:
        np.random.seed(seed)

    if unit_ms:
        t_lim = (t_lim[0]/1000., t_lim[1]/1000.)

    if rate > 0.:
        events_ = [t_lim[0]]

        while events_[-1] < t_lim[1]:
            T = t_lim[1] - events_[-1]

            # expected number of events
            num_expected = T * rate

            # number of events to generate
            num_generate = np.ceil(num_expected+3*np.sqrt(num_expected))
            num_generate = int(max(num_generate, 1000))

            beta = 1. / rate
            isi_ = np.random.exponential(beta, size=num_generate)
            newevents_ = np.cumsum(isi_) + events_[-1]
            events_ = np.append(events_, newevents_)

        lastind = np.searchsorted(events_, t_lim[1])
        events_ = events_[1:lastind]  # drop ghost spike at start

        if unit_ms:
            events_ *= 1000.

    elif rate == 0.:
        events_ = np.asarray([])

    else:
        raise ValueError('requested negative rate.')

    if round_to is not None:
        events_ = np.round(events_/round_to) * round_to

    if seed:
        np.random.seed()

    return events_


def generate_stimulus_spatiotemp(num_pattern, num_inputs, stim_len, t_sim, jitter=5., rate=20., dt=.1, seed=11):
    """
    Generate input and targets for a classification task using spaced input presentations.

    inputs:
        num_pattern         number of base pattern templates
        num_inputs          number of input neurons
        stim_len            length of each pattern
        t_sim               overall time
        jitter              jitter to add to base templates for each
                            presentation
        rate                firing rate of each neuron
        dt                  simulation time step
        seed                seed for pattern generation

    returns:
        spikes              spike input for task
        targets             target value of each presentation
        templates           base templates
    """

    np.random.seed(seed)

    # generate templates
    temp_ = []

    for ns in range(num_pattern):
        temp = [poisson_generator(rate, (0., stim_len), unit_ms=True) for ni in range(num_inputs)]
        temp_ += [temp]

    num_stim = np.floor(t_sim/stim_len).astype(int)

    targets = np.random.randint(num_pattern, size=num_stim)
    start_time = np.arange(num_stim) * stim_len

    spikes_ = [[] for ni in range(num_inputs)]

    for st, target in zip(start_time, targets):
        temp = temp_[target]

        r_ = jitter * np.random.randn(len(temp))

        for ni, (tempi, r) in enumerate(zip(temp, r_)):
            spikes_[ni] += [s + st + r for s in tempi]

    # round

    spikes_round_ = []

    for spikes in spikes_:
        spikes_round = np.round(np.asarray(spikes) / dt) * dt
        spikes_clip = np.unique(np.clip(spikes_round, dt, t_sim))
        spikes_round_ += [spikes_clip.tolist()]

    np.random.seed()

    return spikes_round_, targets, temp_


def generate_stimulus_spatiotemp_multitarget(*args, **kwargs):
    """
    Generate input and targets for three task using spaced input presentations:
    classification, memory, and delayed XOR.

    inputs:
        num_pattern         number of base pattern templates
        num_inputs          number of input neurons
        stim_len            length of each pattern
        t_sim               overall time
        jitter              jitter to add to base templates for each
                            presentation
        rate                firing rate of each neuron
        dt                  simulation time step
        seed                seed for pattern generation

    returns:
        targets             dictionary containing target values for each task
        templates           base templates
        spikes              spike input for task
    """
    spikes, targets, templates = generate_stimulus_spatiotemp(*args, **kwargs)

    num_mem = 3

    targets_ = {
            'classification': targets,
            'memory': {
                k: np.pad(targets, (k, 0), mode='constant')[:-k] for k in range(1, num_mem+1)},
            'xor': np.logical_xor(targets%2, np.pad(targets, (1, 0), mode='constant')[:-1]%2).astype(int),
    }

    return targets_, templates, spikes
