#!/usr/bin/env python3

from joblib import Parallel, delayed
import multiprocessing as mp
import numpy as np


def create_spike_time_grid(times, senders, s_lim, t_lim, *, dt=0.1, drop_silent=False):
    """
    Create a time grid from spike times.

    args:
        times               list of spike times
        senders             list of spike sender ids
        s_lim               tuple containing min and max sender id
        t_lim               tuple containing start and end time of grid

    keyword-only args:
        dt                  discrete time step of output grid, default: 0.1
        drop_silent         don't create a trace for senders without spikes,
                            default: False
    """

    times = np.asarray(times)
    senders = np.asarray(senders)

    assert(s_lim[1] >= s_lim[0]), 's_lim not ordered properly'
    assert(t_lim[1] > t_lim[0]), 't_lim not ordered properly'

    # drop spikes outside of range
    mask = (times >= t_lim[0]) & (times < t_lim[1])
    times = times[mask]
    senders = senders[mask]

    # drop senders outside of range
    mask = (senders >= s_lim[0]) & (senders <= s_lim[1])
    senders = senders[mask]
    times = times[mask]

    assert(len(times) == 0 or min(times) >= t_lim[0])
    assert(len(times) == 0 or max(times) < t_lim[1])
    assert(len(senders) == 0 or min(senders) >= s_lim[0])
    assert(len(senders) == 0 or max(senders) <= s_lim[1])

    # creating time vector as simple index count and changing to actual time
    # values after assignments are done to avoid numerical problems
    delta_t = t_lim[1] - t_lim[0]
    # round may cause spikes to fall outside range, add 1
    L = np.ceil(delta_t / dt).astype(np.int64) + 1
    ti = np.arange(L)

    # create spike grid
    N = s_lim[1] - s_lim[0] + 1
    spike_grid = np.zeros((L, N), dtype=int)

    # create assignment vectors
    times_ind = np.round((times - t_lim[0]) / dt).astype(np.int64)
    senders_ind = (senders - s_lim[0]).astype(np.int64)

    assert(len(times_ind) == 0 or max(times_ind) < L)
    assert(len(senders_ind) == 0 or max(senders_ind) < N)

    # assign
    for time_ind, sender_ind in zip(times_ind, senders_ind):
        spike_grid[time_ind,sender_ind] += 1

    # assert that no spike got lost
    senders_in, spike_counts_in = np.unique(senders, return_counts=True)
    spike_counts_out = spike_grid.sum(axis=0)
    for s, c in zip(senders_in, spike_counts_in):
        assert(spike_counts_out[s-s_lim[0]] == c)

    # remove rows for neurons which never spike if requested
    if drop_silent:
        mask = spike_counts_out > 0
        spike_grid = spike_grid[:,mask]

    # convert time indices to actual time
    t = t_lim[0] + ti * dt

    return t, spike_grid


def lp_filter_spike_train(times, senders, s_lim, t_lim, tau_filter=20., len_filter=100., *, dt=0.1, drop_silent=False):
    """
    Create lowpass-filtered traces from spike events.

    args:
        times               list of spike times
        senders             list of spike sender ids
        s_lim               tuple containing min and mix sender id
        t_lim               tuple containing start and end time of grid
        tau_filter          time constant for exponential filter, default: 20.
        len_filter          length of exponential filter, default: 100.

    keyword-only args:
        dt                  discrete time step of output grid, default: 0.1
        drop_silent         don't create a trace for senders without spikes,
                            default: False
    """

    t, spike_grid = create_spike_time_grid(
            times,
            senders,
            s_lim,
            t_lim,
            dt=dt,
            drop_silent=drop_silent)

    L = spike_grid.shape[0]
    T = L * dt

    # filter with exponential
    f_t = np.arange(0., min(100., T), dt)  # only make as long as necessary
    f = np.exp(-f_t / tau_filter)

    spike_counts = spike_grid.sum(axis=0)
    spikes_filtered = np.zeros(spike_grid.shape)

    for k in range(spike_grid.shape[1]):
        if spike_counts[k] == 0:
            continue

        spikes_filtered[:,k] = np.convolve(spike_grid[:,k], f)[:L]

    return t, spikes_filtered


def __get_liquid_states_worker(neuron_spikes_, times_, tau):
    t_window = 3. * tau
    states_ = np.zeros_like(times_)

    for nt, t in list(enumerate(times_))[::-1]:
        # advance mask
        neuron_spikes_ = neuron_spikes_[neuron_spikes_ < t]

        # only use spikes within window
        spikes_cur_ = neuron_spikes_[neuron_spikes_ >= t - t_window]
        states_[nt] = np.exp(-(t-spikes_cur_) / tau).sum()

    return states_


def get_liquid_states(spikes_, times_, tau, use_mp=True):
    """
    Compute liquid states for given spikes of multiple neurons.

    args:
        spikes_             list of lists containing spike times for a neuron
        times_              times at which to extract liquid states
        tau                 exponential decay time constant
        use_mp              whether to use mulitprocessing, default: True
    """

    N_neuron = len(spikes_)
    N_times = len(times_)
    states_ = np.zeros((N_times, N_neuron))

    if not use_mp:
        ret_ = [__get_liquid_states_worker(ns_, times_, tau) for ns_ in spikes_]

    else:
        delayed_ = (delayed(__get_liquid_states_worker)(ns_, times_, tau) for ns_ in spikes_)
        ret_ = Parallel(n_jobs=mp.cpu_count())(delayed_)

    states_ = np.asarray(ret_).T
    assert states_.shape == (N_times, N_neuron)

    return states_


def spike_hist(times, num_neuron, *, t_lim=None, unit_ms=True, bin_size=10.):
    """
    Compute the mean firing rate of a population over time bins.

    args:
        times               spike times
        num_neuron          number of neurons in input

    keyword-only args:
        t_lim               tuple of start and end time, if None (default),
                            time of first and last spike is used
        bin_size            bin size for rate computation

    returns:
        times               vector of bin center times
        rates               vector of firing rates
    """

    times = np.asarray(times)

    dt = bin_size / 2.

    if t_lim is None:
        t0, t1 = min(times), max(times)
    else:
        t0, t1 = t_lim

    t_ = np.arange(t0, t1+dt, dt)
    t_bin0_, t_bin1_ = t_-dt, t_+dt

    # TODO: more efficient implementation, e.g. with sorted array

    spike_count_ = []

    for t_bin0, t_bin1 in zip(t_bin0_, t_bin1_):
        spike_count_ += [((t_bin0 < times) & (times <= t_bin1)).sum()]

    rate_ = np.asarray(spike_count_) / num_neuron / bin_size
    
    if unit_ms:
        rate_ *= 1000

    return t_, rate_
